#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "clustalw.h"

/*
 *   Prototypes
 */
static Boolean commentline(char *line);
static MRULE * read_user_rules(FILE *fd);
static sint get_rules(MRULE *rules, sint **matrix,sint scale);
static Boolean pattcmp(MRULE rule,DCHAR a,DCHAR b);

/*
 *   Global variables
 */

extern sint	nalphabet;
extern DCHAR 	*alphabet;
extern sint 	gap_pos1, gap_pos2;
extern sint 	max_aa;
extern sint 	mat_avscore;
extern sint 	debug;

DCHAR *codes=NULL;    /* this is a temporary store for the matrix codes */
sint maxres;                 /* they will be x-referenced to alphabet in get_matrix() */
Boolean rules;	/* TRUE if matrix file contains 'rules' */
sint nrules;	/* number of rules read from file */


sint get_matrix(void *mptr, sint **matrix, Boolean neg_flag, sint scale)
{
   sint gg_score = 1;
   sint gr_score = 0;
   sint i, j, k, ix = 0;
   sint av1,av2,av3,min, max;
   DCHAR c1,c2;
   sint *xref;
   Boolean found;
   short *matptr;
   MRULE *ruleptr;
   char line1[MAXLINE+1];


/*
   default - set all scores to 0
*/
   for (i=0;i<=max_aa;i++)
      for (j=0;j<=max_aa;j++)
          matrix[i][j] = 0;

   if (mptr==NULL) {   /* use the identity matrix */
      for (i=0;i<=max_aa;i++) {
		matrix[i][i]=10*scale;
	    matrix[i][gap_pos1]=gr_score;
	    matrix[gap_pos1][i]=gr_score;
	    matrix[i][gap_pos2]=gr_score;
	    matrix[gap_pos2][i]=gr_score;
	    }
	  matrix[gap_pos1][gap_pos1]=gg_score;
	  matrix[gap_pos1][gap_pos2]=gg_score;
	  matrix[gap_pos2][gap_pos1]=gg_score;
	  matrix[gap_pos2][gap_pos2]=gg_score;
      return max_aa;
   }
   else if (rules==TRUE) {
   	/* read rules from ruleptr to matrix */
   	ruleptr=(MRULE *)mptr;
   	maxres=get_rules(ruleptr,matrix,scale);
   }
   else {
/* otherwise, get the matrix from the user file - xref to alphabet */
	if(nalphabet>1) {
		error("Cannot read blast matrix for alphabet length = %d\n",nalphabet);
		return -1;
	}
	matptr=(short *)mptr;
/* first check that all residues used in the sequences are found in the matrix */
   for (i=0;i<max_aa;i++)
     {
	 found=FALSE;
         c1=cpchar(alphabet[i]);
         for (j=0;j<maxres;j++)
           {
           c2=cpchar(codes[j]);
           if (cmpchar(c1,c2))
               {
                  found=TRUE;
                  break;
               }
	   }
         if (found==FALSE)
            {
		strcpy(line1,"residue ");
                strcat(line1,alphabet[i].c);
                strcat(line1," not found in user matrix");
		error(line1);
		return 0;
            }
     }

/*
   cross-reference the matrix residues
*/
   xref=(sint *)ckalloc((maxres+2)*sizeof(sint));
   for (i=0;i<maxres;i++) xref[i] = -1;

   for (i=0;i<maxres;i++)
     {
         c1=cpchar(codes[i]);
         for (j=0;j<=max_aa;j++)
           {
           c2=cpchar(alphabet[j]);
           if (cmpchar(c1,c2))
               {
                  xref[i] = j;
                  break;
               }
	   }
     }



   ix = 0;
   for (i=0;i<maxres;i++)
    {
      for (j=0;j<=i;j++)
       {
               k = matptr[ix];
	       if(xref[i]!=-1 && xref[j]!=-1)
		{
               if (i==j)
                  {
                     matrix[xref[i]][xref[i]] = k * scale;
                  }
               else
                  {
                     matrix[xref[i]][xref[j]] = k * scale;
                     matrix[xref[j]][xref[i]] = k * scale;
                  }
		}
               ix++;
       }
    }
    ckfree(xref);
   }
   
   if (maxres<=0) return -1;
   av1 = av2 = av3 = 0;
   for (i=0;i<=maxres;i++)
    {
      for (j=0;j<=i;j++)
       {
           av1 += matrix[i][j];
           if (i==j)
              {
                 av2 += matrix[i][j];
              }
           else
              {
                 av3 += matrix[i][j];
              }
       }
    }

   av1 /= (maxres*maxres)/2;
   av2 /= maxres;
   av3 /= ((float)(maxres*maxres-maxres))/2;
if (debug>1) fprintf(stdout,"average mismatch score %d\n",(pint)av3);
if (debug>1) fprintf(stdout,"average match score %d\n",(pint)av2);
if (debug>1) fprintf(stdout,"average score %d\n",(pint)av1);

  min = max = matrix[0][0];
  for (i=0;i<=max_aa;i++)
    for (j=1;j<=i;j++)
      {
        if (matrix[i][j] < min) min = matrix[i][j];
        if (matrix[i][j] > max) max = matrix[i][j];
      }
/*
   if requested, make a positive matrix - add -(lowest score) to every entry
*/
  if (neg_flag == FALSE)
   {

if (debug>1) fprintf(stdout,"min %d max %d\n",(pint)min,(pint)max);
      if (min < 0)
        {
           for (i=0;i<=max_aa;i++)
            {
                 for (j=0;j<=max_aa;j++)
                   {
                     matrix[i][j] -= min;
                   }
            }
        }
/*
       gr_score = av3;
       gg_score = -av3;
*/

   }

  mat_avscore = -av3;


  for (i=0;i<gap_pos1;i++)
   {
      matrix[i][gap_pos1] = gr_score;
      matrix[gap_pos1][i] = gr_score;
      matrix[i][gap_pos2] = gr_score;
      matrix[gap_pos2][i] = gr_score;
   }
  matrix[gap_pos1][gap_pos1] = gg_score;
  matrix[gap_pos2][gap_pos2] = gg_score;
  matrix[gap_pos2][gap_pos1] = gg_score;
  matrix[gap_pos2][gap_pos1] = gg_score;

  return(maxres);
}



void * read_user_matrix(char *filename)
{
   short *usermat;
   MRULE *userrules;
   double f;
   FILE *fd;
   sint  numargs,farg;
   sint n, i, j, k = 0;
   char inline1[1024];
   char **args;
   sint ix1, ix = 0;
   float scale;

   if (filename[0] == '\0')
     {
        error("comparison matrix not specified");
       	return((sint)0);
     }

   if ((fd=fopen(filename,"r"))==NULL) 
   {
       	error("cannot open %s", filename);
       	return((sint)0);
   }

   while (fgets(inline1,1024,fd) != NULL) 
	if (!commentline(inline1)) break;
	
   if(strncmp(inline1,"$matrix",7)==0) {
   	rules=TRUE;
   	userrules=read_user_rules(fd);
  	return((void *)userrules);
   }
   
   rules=FALSE;
    
   if(nalphabet>1) {
	error("blast matrix %s is not compatible with alphabet size %d\n",filename,nalphabet);
	fclose(fd);
	return((sint)0);
   }
   n=(sint)pow((double)26,(double)nalphabet);
   args=(char **)ckalloc((n+4) * sizeof(char *));
   if(codes != NULL) ckfree(codes);
   codes=(DCHAR *)ckalloc(n * sizeof(DCHAR));

   rewind(fd);
   maxres = 0;
   while (fgets(inline1,1024,fd) != NULL)
     {
        if (commentline(inline1)) continue;
/*
   read residue characters.
*/
        i = k = 0;
        for (j=0;j<strlen(inline1);)
          {
             if (isalpha((int)inline1[j])) 
		{
		   codes[k].c[i] = inline1[j];
		   if(i>=nalphabet-1) 
			{
			i=0;
		   	k++;
			}
		   else i++;
		   j++;
		}
	     j++;
          }
        codes[k].c[0] = 0;
        break;
    }

   if (k == 0) 
     {
        error("wrong format in matrix %s",filename);
  	fclose(fd);
        return((sint)0);
     }

   maxres=k;

   usermat=(short *)ckalloc((maxres*maxres)*sizeof(short));
/*
   get the weights
*/
   ix = ix1 = 0;
   while (fgets(inline1,1024,fd) != NULL)
     {
        if (blankline(inline1[0])) continue;
        if (inline1[0] == '#' ||
            inline1[0] == '!') break;
        numargs = getargs(inline1, args, (int)(k+1));
        if (numargs < maxres)
          {
             error("wrong format in matrix %s",filename);
  	     fclose(fd);
             return((sint)0);
          }
        if (isalpha(args[0][0])) farg=1;
        else farg=0;

/* decide whether the matrix values are float or decimal */
	scale=1.0;
	for(i=0;i<strlen(args[farg]);i++)
		if(args[farg][i]=='.')
		{
/* we've found a float value */
			scale=10.0;
			break;
		}

        for (i=0;i<=ix;i++)
          {
                  f = atof(args[i+farg]);
                  usermat[ix1++] = (short)(f*scale);
          }
        ix++;
     }
   /*if (ix != k+1)
     {
        error("wrong format in matrix %s",filename);
  	fclose(fd);
        return((sint)0);
     }*/

  ckfree(args);
  fclose(fd);

  return((void *)usermat);
}

static MRULE * read_user_rules(FILE *fd)
{
	MRULE *rules;
	sint i=0,j,l,n,len;
	float score;
	char inline1[1024];
	char pattern[1024];
	
	nrules=0;
	rules=(MRULE *)ckalloc((MAXRULES)*sizeof(MRULE));
 	while (fgets(inline1,1024,fd) != NULL) {
 		i++;
		if (commentline(inline1)) continue;
 		n=sscanf(inline1,"%s%f",pattern,&score);
 		if(n<2) {
 			error("Wrong format in matrix rule file, line %d\n",i);
 			ckfree(rules);
 			return NULL;
 		}
 /* read pattern */
		rules[nrules].pattern=(char **)ckalloc((strlen(pattern)+1)*sizeof(char *));
		for(j=0,len=0;j<strlen(pattern);) {
			if(pattern[j]=='[') {
				l=0; n=0;
				rules[nrules].pattern[len]=(char *)ckalloc((MAXALPHABET+1)*sizeof(char));
				while(pattern[j]!=']') {
					if(pattern[j]==':') {
						l=0;
					}
					else {
						l++;
					}
					rules[nrules].pattern[len][n]=pattern[j];
					n++; j++;
					continue;
				}
				rules[nrules].pattern[len][n]=']';
				rules[nrules].pattern[len][n+1]='\0';
				len+=l;
			}
			else {
				rules[nrules].pattern[len]=(char *)ckalloc((2)*sizeof(char));
				rules[nrules].pattern[len][0]=pattern[j];
				rules[nrules].pattern[len][1]='\0';
				len++;
			}
 			j++;
		}
		
 		if(nrules==0) {
			if (len!=nalphabet) {
				error("matrix patterns are not the same size as the sequence alphabet\n");
				return 0;
			}
 			rules[0].len=len;
		}
		else if (rules[0].len!=len) {
			error("matrix patterns are not all the same size\n");
			return 0;
		}
 		else 
 			rules[nrules].len=len;
 		rules[nrules].score=score;
 		nrules++;
   	} 
	return rules;
}

static sint get_rules(MRULE *rules,sint **matrix,sint scale)
{
/* matrix is already initialised to 0 for i=0,i<=max_aa,
entries should be in the order of alphabet[] */
	sint r,a,b;

/* first, check that the patterns are the same size as the alphabet */	
	if (rules[0].len!=nalphabet) {
		error("matrix patterns are not the same size as the sequence alphabet\n");
		return 0;
	}
	
	for(a=0;a<=max_aa;a++) {
		for(b=a;b<=max_aa;b++) {
			for(r=0;r<nrules;r++) {
			if (pattcmp(rules[r],alphabet[a],alphabet[b])==1) {
				if(rules[r].score>matrix[a][b]) {
					matrix[a][b]=rules[r].score*scale;
					matrix[b][a]=rules[r].score*scale;
				}
			}
			}
		}
	}

if(debug>0) {
sint i,j;
fprintf(stdout,"alphabet\n");
for(i=0;i<=max_aa;i++) fprintf(stdout,"%s ",alphabet[i].c);
fprintf(stdout,"\n");
for(i=0;i<=max_aa;i++) {
for(j=0;j<=max_aa;j++)
 fprintf(stdout,"%4d ",matrix[i][j]);
 fprintf(stdout,"\n");
}
}
	return max_aa;

}

static Boolean pattcmp(MRULE r,DCHAR a,DCHAR b)
{
	sint i,j,k,n,n1;
	Boolean ret=TRUE;
	Boolean found,match;
	char list[10][MAXALPHABET];

	for(i=0;i<r.len;i++) {
		switch (r.pattern[i][0]) {
		case '$': if(a.c[i]!=b.c[i]) return FALSE;
			break;
		case '?':
			break;
		case '[':
			n=0; n1=0;
			for(k=1;r.pattern[i][k]!=']';k++) {
				if(r.pattern[i][k]==':') {
					n++; n1=0;
				}
				else {
					list[n][n1]=r.pattern[i][k];
					n1++;
				}
			}
			n++;
			match=FALSE;
			for(j=0;j<n;j++) {
				found=TRUE;
				for(k=0;k<n1;k++) {
					if (a.c[i+k]!=list[j][k]) {
						found=FALSE;
						break;
					}
				}
				if(found==TRUE) {
					match=TRUE;
					break;
				}
			}
			if(match==FALSE) return FALSE;
			match=FALSE;
			for(j=0;j<n;j++) {
				found=TRUE;
				for(k=0;k<n1;k++) {
					if (b.c[i+k]!=list[j][k]) {
						found=FALSE;
						break;
					}
				}
				if(found==TRUE) {
					match=TRUE;
					break;
				}
			}
			if(match==FALSE) return FALSE;
			i+=n1-1;
 			break;
		default:
			if(a.c[i]!=r.pattern[i][0] || b.c[i]!=r.pattern[i][0])
				return FALSE;
		}
	}
	return ret;
}

int getargs(char *inline1,char *args[],int max)
{

	char	*inptr;
/*
#ifndef MAC
	char	*strtok(char *s1, const char *s2);
#endif
*/
	int	i;

	inptr=inline1;
	for (i=0;i<=max;i++)
	{
		if ((args[i]=strtok(inptr," \t\n"))==NULL)
			break;
		inptr=NULL;
	}

	return(i);
}


static Boolean commentline(char *line)
{
        int i;
 
        if(line[0] == '#') return TRUE;
        for(i=0;line[i]!='\n' && line[i]!=EOS;i++) {
                if(!isspace(line[i]))
			return FALSE;
        }
        return TRUE;
}

sint matrix_score(sint **matrix,sint c1,sint c2)
{
        sint ret;

        ret=matrix[c1][c2];

	return ret;
}

